# Copyright (C) 2005 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: dmpdb.py,v 1.3 2005/08/14 19:55:58 dieter Exp $
'''Dieter Maurers PDB extension.

'PDB', the Python debugger, has severe limitations (apart from some bugs).
This extension tries to get rid of some of them.

What we plan to change finally:

 * setting breakpoints externally

   Currently, we must be inside the debugger to set breakpoints

   We want to be able to set breakpoints directly on the debugger
   object without the need to enter the debugger

   Implemented.

 * exception display

   The current display of an exception is almost useless.

   We plan to drastically improve on this: e.g. allow
   to view the full exception information, including a traceback.
   The details are not yet known.

   This needs to be better understood.
   Looks as if "Exception.__repr__" and "Exception.__str__" were
   incredibly stupid.

 * 'where' command improvements

   - each frame identifies its level (the top frame is at level 0,
     the one below it at level 1, etc...)

   - the current frame is marked as such

   - 'where' gets optional arguments *number* and *end*

     *number* controls how many frames are displayed (default: 'maxint'),
     *end* at what level the display ends, negative numbers count from
     the bottom (default: '-1', i.e. the bottom frame).

   Implemented.

 * new command 'frame level' (abbr: 'f')

   switch to call frame at *level*

   without argument, provide info about current frame

   Implemented

 * allow to customize the display of the 'where' command
   and of tracebacks

   e.g. to display additional debugging information as
   provided by Zope (for example)
'''

from pdb import Pdb, set_trace, post_mortem, pm
import sys

from ReuseUtils import rebindFunction


class _CurframeRedirector(object):
  '''auxiliary class to define the 'curframe' descriptor.'''
  def __get__(self, instance, owner):
    if instance is None: raise AttributeError('curframe')
    cf = getattr(instance, '_curframe', self)
    if cf is self:
      # curframe is undefined; use that of the caller
      # Note: that we should maybe advance until we leave "pdb".
      cf = sys._getframe(2)
    return cf

  def __set__(self, instance, value):
    setattr(instance, '_curframe', value)

  def __delete__(self, instance):
    delattr(instance, '_curframe')
  

class Pdb(Pdb, object):
  '''PDB extension.'''

  curframe = _CurframeRedirector()

  def print_stack_entry(self, frame_lineno, prompt_prefix=None, frameno=None):
    if prompt_prefix is None:
      if frameno is None: frameno = self.curindex
      prompt_prefix = '\n  [%d] ' % frameno
    super(Pdb, self).print_stack_entry(frame_lineno, prompt_prefix)
    info = self.getAdditionalFrameInfo(frame_lineno)
    if info:
      if isinstance(info, (str, unicode)): info = info.split('\n')
      print '  ' + '\n  '.join(info)

  def print_stack_trace(self, number=sys.maxint, end=-1):
    if end < 0: end += len(self.stack)
    if end < 0: return
    start = end+1 - number
    if start < 0: start = 0
    try:
      for index in range(start, end+1):
        frame_lineno = self.stack[index]
        self.print_stack_entry(frame_lineno, frameno=index)
    except KeyboardInterrupt: pass

  def do_where(self, arg):
    arg = arg.strip()
    number = sys.maxint; end = -1
    if arg:
      args = arg.split()
      number = int(args[0])
      if len(args) > 1: end = int(args[1])
    self.print_stack_trace(number, end)
  do_w = do_bt = do_where

  def do_frame(self, arg):
    arg = arg.strip()
    if arg: arg = int(arg)
    else: arg = self.curindex
    if arg < 0: arg += len(self.stack)
    if not (0 <= arg < len(self.stack)):
      print '*** no such frame'
      return
    elif arg != self.curindex:
      self.curindex = arg
      self.curframe = self.stack[arg][0]
      self.lineno = None
    self.print_stack_entry(self.stack[arg])
  do_f = do_frame

  def help_where(self):
    self.help_w()

  def help_w(self):
    print """w(here)
Print a stack trace, with the most recent frame at the bottom.
An arrow indicates the "current frame", which determines the
context of most commands.  'bt' is an alias for this command.

It supports optional arguments "number" and "end".
"number" specifies the number of frames to be printed (default: maxint);
"end" specifies the last frame to be printed, a negative value counts
from the end (default: "-1", i.e. the bottom frame)."""

  help_bt = help_w

  def help_frame(self):
    self.help_f()

  def help_f(self):
    print """f(rame)
With an argument (an integer), first switch to that frame (a negative value
counts from the end).
In any case, print (then) information about the current frame."""

  def getAdditionalFrameInfo(self, frame_lineno):
    '''returns additional info, if available.

    If not a false value, this should either be a string or a sequence

    *frame_lineno* is a tuple *(frame, lineno)*.
    '''
    return
    
set_trace = rebindFunction(set_trace, Pdb=Pdb)
post_mortem = rebindFunction(post_mortem, Pdb=Pdb)
pm = rebindFunction(pm, post_mortem=post_mortem)
